"""Bakery house."""

from typing import Optional

from .stuff import BakeryLogger, DefaultLogger


logger: Optional[BakeryLogger] = DefaultLogger()

from .bakery import *
from .baking import *
from .cake import *
from .piece_of_cake import *
from .stuff import *


__all__ = [
    *bakery.__all__,  # type: ignore # pylint: disable=undefined-variable
    *baking.__all__,  # type: ignore # pylint: disable=undefined-variable
    *cake.__all__,  # type: ignore # pylint: disable=undefined-variable
    *piece_of_cake.__all__,  # type: ignore # pylint: disable=undefined-variable
    *stuff.__all__,  # type: ignore # pylint: disable=undefined-variable
]



"""Cake.

Book, recipes, etc.
"""

__all__ = ["Cake", "Pastry", "hand_made"]

from typing import (
    Any,
    AsyncContextManager,
    Awaitable,
    Callable,
    ContextManager,
    Optional,
    Set,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Final, ParamSpec

from .baking import Ingredients
from .piece_of_cake import PieceOfCake
from .stuff import (
    BakingMethod,
    Cakeable,
    CakeRecipe,
    IngredientsProto,
    assert_baked,
    cake_ingredients,
    is_baked,
    is_cake,
)


class Pastry(CakeRecipe):
    """Pastry is a public cake interface with almost zero name collision.

    Your item ingredients and cooking method stored here.
    """

    __CAKE_ATTRIBUTE_ERROR_NAMES__: Final[Set[str]] = {
        "func",  # partial
        "__wrapped__",  # functools special attribute
    }

    def __init__(
        self,
        ingredients: Ingredients,
    ):
        self.__ingredients: Final[Ingredients] = ingredients
        self.__name__: Final = ingredients.__name__
        self.__code__: Final = ingredients.__code__
        self.__defaults__: Final = ingredients.__defaults__
        self.__kwdefaults__: Final = ingredients.__kwdefaults__
        self.__annotations__: Final = ingredients.__annotations__
        self._is_coroutine: Final = ingredients._is_coroutine

    def __set_name__(self, _: Any, name: str) -> None:
        self.__ingredients.name = name

    def __repr__(self) -> str:
        """Repr."""
        return self.__ingredients.__repr__()

    def __copy__(self) -> "Cakeable[Any]":
        """Copy itself with all ingredients and technologies.

        If cake is not baked then such a cake's copy is cake itself.

        'Copy cake' just sounds like 'cupcake'. And I like cupcakes ;)
        """
        if not is_baked(self):
            return self

        ingr_copy: Ingredients = self.__ingredients.__copy__()
        return cast(Cakeable[Any], Pastry(ingr_copy))

    def __deepcopy__(self, memo: Any) -> "Cakeable[Any]":
        """Deep copy with all ingredients and technologies.

        If cake is not baked then such a cake's deep copy is cake itself.
        """
        if not is_baked(self):
            return self

        ingr_copy: Ingredients = self.__ingredients.__deepcopy__(memo)
        return cast(Cakeable[Any], Pastry(ingr_copy))

    def __call__(self) -> Any:
        """Just return whole cake."""
        assert_baked(cast(Cakeable[Any], self))
        return self.__ingredients()

    def __getattr__(self, piece_name: str) -> PieceOfCake:
        """Cut a piece of cake.

        With __CAKE_ATTRIBUTE_ERROR_NAMES__ you can no longer write
        the things like:
        class MyBakery(Bakery):
            ctrl: Any = Cake(Controller)
            controller_func: Any = Cake(some_func, ctrl.func)  # or __wrapped__

        There is the place where AttribureError exception will occured.
        You can make a simple refactoring:
        class MyBakery(Bakery):
            ctrl: Any = Cake(Controller)
            controller_func: Any = Cake(some_func, Cake(getattr, ctrl, "func"))
        """
        if piece_name in self.__CAKE_ATTRIBUTE_ERROR_NAMES__:
            raise AttributeError(piece_name)
        # explicit __getattr__ call to avoid collisions
        return PieceOfCake(self).__getattr__(piece_name)

    def __getitem__(self, piece_name: Any) -> PieceOfCake:
        """Cut a piece of cake."""
        return PieceOfCake(self).__getitem__(piece_name)

    async def __aenter__(self) -> Any:
        return await self.__ingredients.bake()

    async def __aexit__(
        self,
        exc_type: Optional[type],
        exc_value: Optional[Exception],
        traceback: Optional[Any],
    ) -> None:
        return await self.__ingredients.unbake(exc_type, exc_value, traceback)


T = TypeVar("T")
P = ParamSpec("P")


def hand_made(cake: T, cake_baking_method: BakingMethod) -> T:
    """Hand made cake."""

    if not is_cake(cake):
        cake = Cake(cake)

    ingredients: IngredientsProto = cake_ingredients(cast(Cakeable[Any], cake))
    ingredients.cake_baking_method = cake_baking_method
    return cake


# pylint: disable=invalid-name


@overload
def Cake(recipe: Awaitable[T]) -> T:
    ...


@overload
def Cake(recipe: AsyncContextManager[T]) -> T:
    ...


@overload
def Cake(recipe: ContextManager[T]) -> T:
    ...


@overload
def Cake(
    recipe: Callable[P, Awaitable[T]],
    *recipe_args: P.args,
    **recipe_kwargs: P.kwargs,
) -> T:
    ...


@overload
def Cake(
    recipe: Callable[P, T],
    *recipe_args: P.args,
    **recipe_kwargs: P.kwargs,
) -> T:
    ...


@overload
def Cake(recipe: T) -> T:
    ...


def Cake(  # waiting for issue to close
    # https://github.com/python/mypy/issues/11004
    recipe: T,
    *recipe_args: Any,
    **recipe_kwargs: Any,
) -> T:
    """Cake as Cake =)."""
    return cast(
        T,
        Pastry(Ingredients(recipe, *recipe_args, **recipe_kwargs)),
    )



"""Empty cake shape.
For future purposes (maybe).

Goal: Pass arguments to dependent bakery.
case 1: # <-----
Main bakery defines settings. Only once, it's required.
Also main bakery has dependent bakery.
So, main bakery passes some settings to dependent bakery.
case2:  # <-----
Main bakery initializes dependent bakery 1 and dependent bakery 2.
So from dependent bakery 1 pass some cakes to dependent bakery 2.
Perfect impl:
    class MyBakery(Bakery):
        some_field: int
        ya_field: str
        my_class: Cake = Cake(MyClass, some_field, ya_field)
"""


# __all__ = ["Shape"]
#
# from typing import Any, Generic, Optional, Protocol, TypeVar, cast
#
# from .piece_of_cake import PieceOfCake
# from .stuff import CakeRecipe
#
#
# T = TypeVar("T")
#
#
# class CakeType(Protocol):
#    """cake protocol."""
#
#    def __call__(self):
#        ...
#
#    def __getattr__(self, piece_name: str) -> Any:
#        ...
#
#    def __getitem__(self, piece_name: Any) -> Any:
#        ...
#
#    async def __aenter__(self) -> Any:
#        ...
#
#    async def __aexit__(self, *_args: Any) -> None:
#        ...
#
#    async def bake(self) -> Any:
#        """bake."""
#
#    async def unbake(self, *_args: Any) -> None:
#        """unbake."""
#
#    def __set_name__(self, _, name: str) -> None:
#        ...
#
#
# class ShapeType(Protocol):
#    """shape protocol."""
#
#    _cake: Optional[CakeType]
#
#
# def assert_filled(shape: ShapeType) -> None:
#    """assert shape is filled."""
#    if shape._cake is None:
#        raise ValueError(f"{shape} is not filled. Just fill it with any cake!")
#
#
# def casted_cake(shape: ShapeType) -> CakeType:
#    """assert shape is filled and return shape cake."""
#    assert_filled(shape)
#    return cast(CakeType, shape._cake)
#
#
# class Shape(CakeRecipe, Generic[T]):
#    """Shape for baking."""
#
#    def __init__(self):
#        self._name: str = ""
#        self._cake: Optional[CakeType] = None
#
#    def __set_name__(self, _, name: str) -> None:
#        """shape name."""
#        self._name = name
#
#    def __repr__(self) -> str:
#        """repr."""
#        name: str = self._name or "<anon>"
#        return f"Shape '{name}'"
#
#    def __call__(self) -> T:
#        return casted_cake(self).__call__()
#
#    def __getattr__(self, piece_name: str) -> PieceOfCake:
#        return PieceOfCake(self).__getattr__(piece_name)
#
#    def __getitem__(self, piece_name: str) -> PieceOfCake:
#        return PieceOfCake(self).__getitem__(piece_name)
#
#    async def __aenter__(self) -> Any:
#        return await self.bake()
#
#    async def __aexit__(self, *args: Any) -> Any:
#        return await self.unbake(*args)
#
#    async def bake(self) -> Any:
#        """bake cake in shape."""
#        return await casted_cake(self).bake()
#
#    async def unbake(self, *args: Any) -> None:
#        """unbake cake in shape."""
#        return await casted_cake(self).unbake(*args)
#
#    def __cake(self, cake: CakeType) -> None:
#        """set cake."""
#        if self._cake is not None:
#            raise ValueError(f"{self} is already filled.")
#        if not isinstance(cake, CakeRecipe):
#            raise ValueError(f"You only can fill {self} with a cake, not {type(cake)}")
#        cake.__set_name__(self, self._name)
#        self._cake = cake
#
#    cake = property(fget=None, fset=__cake)



"""Piece of any cake."""

__all__ = [
    "PieceAttr",
    "PieceOfCake",
    "PieceSubs",
    "PieceType",
]

from typing import Any, List

from typing_extensions import Final

from .stuff import Cakeable, FictionalPiece, is_cake


class PieceType:
    """Piece type."""

    def __init__(self, mark: Any):
        self.mark: Any = mark


class PieceAttr(PieceType):
    """Piece attribute."""


class PieceSubs(PieceType):
    """Piece subscription."""


def copy_piece_of_cake(to_copy: "PieceOfCake", mark: PieceType) -> "PieceOfCake":
    """Copy piece_of_cake with mark."""
    piece_copy: PieceOfCake = PieceOfCake(to_copy.cake)
    piece_copy.pieces.extend([*to_copy.pieces, mark])
    return piece_copy


class PieceOfCake(FictionalPiece):
    """Any piece of your any cake."""

    def __init__(self, cake: Cakeable[Any]):
        self.cake: Final[Cakeable[Any]] = cake
        self.pieces: Final[List[PieceType]] = []

    def __repr__(self) -> str:
        res: str = str(self.cake)
        for piece in self.pieces:
            piece_repr: str = f".{piece.mark}"
            if isinstance(piece, PieceSubs):
                piece_repr = f"[{piece.mark}]"
            res += piece_repr

        return res

    def __iter__(self) -> None:
        raise ValueError("Piece of cake is not iterable")

    def __getattr__(self, mark: Any) -> "PieceOfCake":
        """Remember all marks were done."""
        return copy_piece_of_cake(self, PieceAttr(mark))

    def __getitem__(self, mark: Any) -> "PieceOfCake":
        """Remember all subscriptions."""
        return copy_piece_of_cake(self, PieceSubs(mark))

    def __call__(self, *_args: Any, **_kwargs: Any) -> Any:
        """It's time to cut a piece from cake.
        Get dummy *args and **kwargs to highlight
        `Cake is not baked` errors.
        Example:
        ```code
            database: Cake
            await database.write_values([1,2,3,4,5], table="some_table")
            # Without dummy args some strange error occured:
            # `PieceOfCake.__call__() takes 1 positional argument but ... were given`.
        ```
        """

        cake: Cakeable[Any] = self.cake
        if is_cake(cake):
            cake = cake()

        for piece in self.pieces:
            mark: Any = piece.mark
            if mark.__class__ is self.__class__:
                mark = mark()

            if isinstance(piece, PieceAttr):
                cake = getattr(cake, mark)
            elif isinstance(piece, PieceSubs):
                cake = cake[mark]  # type: ignore
            else:
                raise ValueError(f"Unknown piece '{piece}' of cake '{self.cake}'")

        if is_cake(cake):
            cake = cake()
        return cake



"""Baking stuff."""

__all__ = [
    "BUILTIN_TYPES",
    "BakeryLogger",
    "BakingMethod",
    "CakeRecipe",
    "Cakeable",
    "DefaultLogger",
    "IngredientsProto",
    "cake_baking_method",
    "cake_ingredients",
    "cake_name",
    "flatten",
    "is_baked",
    "is_cake",
    "is_iterable",
    "is_mapping",
    "is_piece_of_cake",
]

import types
from copy import copy
from datetime import datetime
from enum import IntEnum, auto
from functools import partial
from inspect import Signature
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Final, Protocol, TypeGuard

import bakery


class BakingMethod(IntEnum):
    """Baking method."""

    BAKE_AUTO = 0
    BAKE_FROM_CORO_FUNC = auto()
    BAKE_FROM_AWAITABLE = auto()
    BAKE_FROM_ACM = auto()
    BAKE_FROM_CM = auto()
    BAKE_FROM_BUILTIN = auto()
    BAKE_FROM_CALL = auto()
    BAKE_NO_BAKE = auto()


class IngredientsProto(Protocol):
    """Ingredients protocol."""

    recipe: Any
    recipe_args: Any
    recipe_kwargs: Any
    cake_baking_method: BakingMethod
    is_baked: bool
    name: str

    def __call__(self) -> Any:
        ...


class CakeRecipe:
    """Any recipe you want."""

    # for fastapi Depends
    __signature__ = Signature()

    def __call__(self) -> Any:
        """Do something."""
        raise NotImplementedError


class FictionalPiece:
    """Fictional piece of cake."""

    # for fastapi Depends
    __signature__ = Signature()

    def __call__(self) -> Any:
        """Do something."""
        raise NotImplementedError


T_co = TypeVar("T_co", covariant=True)


class PieceProto(Protocol):
    """Piece of cake protocol."""

    def __getattr__(self, mark: Any) -> "PieceProto":
        """Getattr."""

    def __getitem__(self, mark: Any) -> "PieceProto":
        """Subscription."""

    def __call__(self) -> Any:
        """Get cake."""


class Cakeable(Protocol[T_co]):
    """Cakeable protocol."""

    def __set_name__(self, _: Any, name: str) -> None:
        """Set cakeable name."""

    def __repr__(self) -> str:
        """Repr."""

    def __copy__(self) -> "Cakeable[T_co]":
        """Copy cake."""

    def __call__(self) -> T_co:
        """Call cake."""

    def __getattr__(self, piece_name: str) -> PieceProto:
        """Getattr."""

    def __getitem__(self, piece_name: Any) -> PieceProto:
        """Subscription."""

    async def __aenter__(self) -> T_co:
        """Bake cakeable."""

    async def __aexit__(
        self,
        exc_type: Optional[type],
        exc_value: Optional[Exception],
        traceback: Optional[Any],
    ) -> None:
        """Unbake cakeable."""


def is_iterable(
    value: Any,
    ignore_types: Tuple[Type[Any], ...] = (
        str,
        bytes,
        memoryview,
        types.GeneratorType,
        types.AsyncGeneratorType,
        map,
        range,
        zip,
        CakeRecipe,
        FictionalPiece,
    ),
) -> bool:
    """Check if iterable, ignore generator-like by default."""
    return isinstance(value, Iterable) and not isinstance(value, ignore_types)


def is_mapping(value: Any) -> TypeGuard[Mapping[Any, Any]]:
    """Is mapping check."""
    return isinstance(value, Mapping)


def is_cake(value: Any) -> bool:
    """Is cake."""
    return isinstance(value, CakeRecipe)


def is_piece_of_cake(value: Any) -> bool:
    """Is piece of cake."""
    return isinstance(value, FictionalPiece)


def is_cake_or_piece(value: Any) -> bool:
    """Is cake or piece of cake."""
    return is_cake(value) or is_piece_of_cake(value)


def cake_ingredients(cake: Cakeable[Any]) -> IngredientsProto:
    """Cake ingredients."""
    if not is_cake(cake):
        raise ValueError(f"Only cakes are baking in the bakery, not {cake}")
    return cast(IngredientsProto, getattr(cake, "_Pastry__ingredients"))


def cake_name(cake: Cakeable[Any]) -> str:
    """Is cake with name."""
    return cake_ingredients(cake).name


def is_baked(cake: Cakeable[Any]) -> bool:
    """Check if recipe is baked."""
    return cake_ingredients(cake).is_baked


def assert_baked(cake: Cakeable[Any]) -> None:
    """Check if recipe is baked."""
    if not is_baked(cake):
        raise ValueError(f"{cake} is not baked. Just bake it!")


def cake_baking_method(cake: Cakeable[Any]) -> IntEnum:
    """Cake baking method."""
    return cake_ingredients(cake).cake_baking_method


def flatten(
    items: Union[Iterable[Any], Mapping[Any, Any]],
    ignore_types: Tuple[Type[Any], ...] = (
        str,
        bytes,
        memoryview,
        types.GeneratorType,
        types.AsyncGeneratorType,
        map,
        range,
        zip,
        CakeRecipe,
        FictionalPiece,
    ),
) -> Iterator[Any]:
    """Flatten items, except ignored (generator-like)."""
    items = items.items() if is_mapping(items) else items
    for _item in items:
        if is_iterable(_item, ignore_types) or is_mapping(_item):
            yield from flatten(_item)
        else:
            yield _item


def replace_cakes(obj: Any) -> Any:
    """Replace all objects."""

    res: Any

    if is_mapping(obj):
        res = copy(obj)
        for key, value in obj.items():
            _key = replace_cakes(key)
            _value = replace_cakes(value)
            res[_key] = _value

    elif is_iterable(obj):
        res = type(obj)(replace_cakes(item) for item in obj)

    elif is_cake_or_piece(obj):
        res = obj()

    else:
        res = obj

    return res


BUILTIN_TYPES: Final[Tuple[Any, ...]] = (
    bool,
    bytearray,
    bytes,
    classmethod,
    complex,
    dict,
    float,
    frozenset,
    int,
    list,
    map,
    memoryview,
    property,
    range,
    set,
    slice,
    staticmethod,
    str,
    super,
    tuple,
    zip,
    type(None),
    type(Ellipsis),
    types.AsyncGeneratorType,
    types.GeneratorType,
    types.ModuleType,
    types.SimpleNamespace,
)


class BakeryLogger(Protocol):
    """Bakery logger."""

    def debug(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Debug."""

    def info(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Info."""

    def warning(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Warning."""

    def error(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Error."""


class DummyLogger:
    """Dummy logger."""

    def debug(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Debug."""

    def info(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Info."""

    def warning(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Warning."""

    def error(self, __message: str, *args: Any, **kwargs: Any) -> None:
        """Error."""


class DefaultLogger:
    """Just instead of print function."""

    FUNC_2_LEVEL: Final[Dict[str, str]] = {
        "debug": "DEBUG",
        "info": "INFO ",
        "warning": "WARN ",
        "error": "ERROR",
    }

    @staticmethod
    def _log(level: str, message: str) -> None:
        """Log it."""
        print(
            f"{datetime.now().isoformat(sep=' ', timespec='milliseconds')} | {level} | {message}"
        )

    def __getattr__(self, attr: str) -> Callable[..., Any]:
        if attr not in self.FUNC_2_LEVEL:
            raise AttributeError(
                f"Logger has no attribute {attr}. Possible values: {list(self.FUNC_2_LEVEL)}"
            )

        return partial(self._log, self.FUNC_2_LEVEL[attr])


DUMMY_LOGGER: Final[DummyLogger] = DummyLogger()


class LoggerWrapper:
    """Logger wrapper."""

    def __getattr__(self, attr: str) -> Any:
        """Return actual logger method."""
        cur_logger: BakeryLogger = bakery.logger or DUMMY_LOGGER
        return getattr(cur_logger, attr)


_LOGGER: Final[LoggerWrapper] = LoggerWrapper()



"""All relative to baking."""


__all__ = [
    "Ingredients",
    "bake",
    "check_baking_method",
    "determine_baking_method",
    "unbake",
]
from copy import deepcopy
from inspect import isawaitable, iscoroutinefunction
from typing import Any, AsyncContextManager, ContextManager, Optional, TypeVar

from typing_extensions import Final

from .stuff import _LOGGER as logger
from .stuff import (
    BUILTIN_TYPES,
    BakingMethod,
    Cakeable,
    cake_name,
    flatten,
    is_baked,
    is_cake,
    is_cake_or_piece,
    is_piece_of_cake,
    replace_cakes,
)


def determine_baking_method(recipe: Any) -> BakingMethod:
    """Determine the first available baking method for recipe."""
    if is_cake_or_piece(recipe):
        # Baking method depends on what object
        # really is inside cake/piece_of_cake
        return BakingMethod.BAKE_AUTO

    for method in [
        BakingMethod.BAKE_FROM_CORO_FUNC,
        BakingMethod.BAKE_FROM_AWAITABLE,
        BakingMethod.BAKE_FROM_ACM,
        BakingMethod.BAKE_FROM_CM,
        BakingMethod.BAKE_FROM_BUILTIN,
        BakingMethod.BAKE_FROM_CALL,
    ]:
        if check_baking_method(recipe, method):
            return method

    logger.info(f"Cannot determine baking method for recipe {recipe}")
    return BakingMethod.BAKE_NO_BAKE


def check_baking_method(recipe: Any, method: BakingMethod) -> bool:
    """Check baking method.

    return true if OK.
    """
    if method == BakingMethod.BAKE_FROM_CALL:
        return callable(recipe)

    if method == BakingMethod.BAKE_FROM_CM:
        return isinstance(recipe, ContextManager)

    if method == BakingMethod.BAKE_FROM_ACM:
        return isinstance(recipe, AsyncContextManager)

    if method == BakingMethod.BAKE_FROM_BUILTIN:
        return isinstance(recipe, BUILTIN_TYPES)

    if method == BakingMethod.BAKE_FROM_CORO_FUNC:
        return iscoroutinefunction(recipe)

    if method == BakingMethod.BAKE_FROM_AWAITABLE:
        return isawaitable(recipe)

    if method == BakingMethod.BAKE_NO_BAKE:
        return True

    raise ValueError(f"Cannot validate unknown baking method '{method}'.")


T = TypeVar("T")


async def bake(cake: Cakeable[T]) -> T:
    """Bake cake."""
    return await cake.__aenter__()


async def unbake(
    cake: Cakeable[Any],
    exc_type: Optional[type] = None,
    exc_value: Optional[Exception] = None,
    traceback: Optional[Any] = None,
) -> None:
    """Unbake."""
    return await cake.__aexit__(exc_type, exc_value, traceback)


class Ingredients:
    """Cake ingredients.

    Cake internal state.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        recipe: Any,
        *recipe_args: Any,
        cake_baking_method: BakingMethod = BakingMethod.BAKE_AUTO,
        **recipe_kwargs: Any,
    ):

        self.recipe: Final[Any] = recipe
        self.recipe_args: Final[Any] = recipe_args
        self.recipe_kwargs: Final[Any] = recipe_kwargs

        self.cake_baking_method: BakingMethod = cake_baking_method

        self.result: Any = None

        self.is_baked: bool = False

        self.name: str = ""

        if not self.cake_baking_method:
            self.cake_baking_method = determine_baking_method(self.recipe)
        else:
            check_baking_method(self.recipe, cake_baking_method)

        if self.cake_baking_method == BakingMethod.BAKE_NO_BAKE:
            self.result = self.recipe
            self.is_baked = True

        self.__name__: Final = getattr(self.recipe, "__name__", None) or self.__call__.__name__
        self.__code__: Final = getattr(self.recipe, "__code__", None) or self.__call__.__code__
        self.__defaults__: Final = (
            getattr(self.recipe, "__defaults__", None) or self.__call__.__defaults__
        )
        self.__kwdefaults__: Final = (
            getattr(self.recipe, "__kwdefaults__", None) or self.__call__.__kwdefaults__
        )
        self.__annotations__: Final = (
            getattr(self.recipe, "__annotations__", None) or self.__call__.__annotations__
        )
        self._is_coroutine: Final = getattr(self.recipe, "_is_coroutine", False)

    def __repr__(self) -> str:
        """Repr."""
        name: str = self.name or "<anon>"
        return f"Cake '{name}'"

    def __copy__(self) -> "Ingredients":
        """Copy all ingredients and technologies."""
        ingr_copy: Ingredients = Ingredients(
            self.recipe,
            *list(self.recipe_args),
            cake_baking_method=self.cake_baking_method,
            **dict(self.recipe_kwargs),
        )
        ingr_copy.name = self.name
        return ingr_copy

    def __deepcopy__(self, memo: Any) -> "Ingredients":
        """Deep copy all ingredients and technologies."""
        ingr_copy: Ingredients = Ingredients(
            self.recipe,
            *deepcopy(self.recipe_args),
            cake_baking_method=self.cake_baking_method,
            **deepcopy(self.recipe_kwargs),
        )
        ingr_copy.name = self.name
        return ingr_copy

    def __call__(self) -> Any:
        """Just return result."""
        return self.result

    async def bake(self) -> Any:
        """Bake it all."""

        if self.is_baked:
            return self.result

        for recipe in flatten([self.recipe_args, self.recipe_kwargs, self.recipe]):
            if is_cake(recipe):
                await bake(recipe)

        recipe = self.recipe
        if is_cake_or_piece(recipe):
            recipe = self.recipe()

        if not self.cake_baking_method:
            self.cake_baking_method = determine_baking_method(recipe)

        if self.cake_baking_method == BakingMethod.BAKE_FROM_BUILTIN:
            self.result = replace_cakes(self.recipe)

        elif self.cake_baking_method == BakingMethod.BAKE_FROM_CALL:
            self.result = recipe(
                *replace_cakes(self.recipe_args),
                **replace_cakes(self.recipe_kwargs),
            )

        elif self.cake_baking_method == BakingMethod.BAKE_FROM_CM:
            self.result = recipe.__enter__()

        elif self.cake_baking_method == BakingMethod.BAKE_FROM_ACM:
            self.result = await recipe.__aenter__()

        elif self.cake_baking_method == BakingMethod.BAKE_FROM_CORO_FUNC:
            self.result = await recipe(
                *replace_cakes(self.recipe_args),
                **replace_cakes(self.recipe_kwargs),
            )

        elif self.cake_baking_method == BakingMethod.BAKE_FROM_AWAITABLE:
            self.result = await recipe

        elif self.cake_baking_method == BakingMethod.BAKE_NO_BAKE:
            self.result = recipe

        else:
            raise ValueError(
                f"{self}: Unknown baking method '{self.cake_baking_method}' "
                f"for recipe {self.recipe}"
            )

        logger.debug(f"{self} is baked [{self.cake_baking_method.name}]")
        self.is_baked = True
        return self.result

    async def unbake(
        self,
        exc_type: Optional[type] = None,
        exc_value: Optional[Exception] = None,
        traceback: Optional[Any] = None,
    ) -> None:
        """Unbake cake."""

        # Unbake anonymous recipes even if not self.is_baked
        # Save recipe called value (recipe())
        # because it will be impossible to do it
        # after recipe unbaked
        recipe: Any = self.recipe

        is_cake_and_baked: bool = is_cake(recipe) and is_baked(recipe)
        is_poc_and_baked: bool = is_piece_of_cake(recipe) and is_baked(recipe.cake)

        if is_cake_and_baked or is_poc_and_baked:
            recipe = recipe()

        _recipe: Any
        for _recipe in flatten([self.recipe, self.recipe_args, self.recipe_kwargs]):
            if is_cake(_recipe) and not cake_name(_recipe):
                # unbake anonymous recipes only
                await unbake(_recipe, exc_type, exc_value, traceback)

        if not self.is_baked:
            return

        # All cakes and piece_of_cakes should already be unbaked
        # at this moment
        if not is_cake_or_piece(recipe):
            if self.cake_baking_method == BakingMethod.BAKE_FROM_CM:
                recipe.__exit__(exc_type, exc_value, traceback)

            elif self.cake_baking_method == BakingMethod.BAKE_FROM_ACM:
                await recipe.__aexit__(exc_type, exc_value, traceback)

        logger.debug(f"{self} is unbaked")

        self.is_baked = False



# isort: skip_file
from typing import Any, Dict, Type, TypeVar

from .stuff import Cakeable

T = TypeVar('T', bound='Bakery')

class Bakery:
    __bakery_visitors__: int
    __bakery_items__: Dict[str, Cakeable[Any]]
    async def __aenter__(self: T) -> T: ...
    async def __aexit__(self, *_args: Any) -> None: ...
    @classmethod
    async def aopen(cls: Type[T]) -> T: ...
    @classmethod
    async def aclose(
        cls,
        *_args: Any,
    ) -> None: ...



"""DI Bakery.

Opened around the clock.
"""

__all__ = ["Bakery"]

from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from .baking import bake, unbake
from .cake import Cake
from .stuff import _LOGGER as logger
from .stuff import Cakeable, is_cake


T = TypeVar('T', bound='Bakery')


class Bakery:
    """Your bakery."""

    __bakery_visitors__: int
    __bakery_items__: Dict[str, Cakeable[Any]]

    async def __aenter__(self: T) -> T:
        """Open up your real bakery."""
        return await type(self).aopen()

    async def __aexit__(self, *_args: Any) -> None:
        """Close up bakery.

        Unbake all cakes.
        """
        return await type(self).aclose()

    def __getattribute__(self, attr: str) -> Any:
        """If getattr(cls, attr) is cake, call it and return."""
        value: Any = super().__getattribute__(attr)
        if is_cake(value):
            return value()
        return value

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize bakery subclass."""
        bakery_items: Dict[str, Cakeable[Any]] = {}
        # Do filter __dict__, because iterating
        # over __annotations__ forces to annotate
        # every cake
        for cake_name, cake in cls.__dict__.items():
            if cake_name.startswith("__") and cake_name.endswith("__"):
                continue
            if not is_cake(cake):
                # wrap value into cake
                cake = Cake(cake)
                setattr(cls, cake_name, cake)
                cake.__set_name__(cls, cake_name)
            bakery_items[cake_name] = cake

        cls.__bakery_items__ = bakery_items
        cls.__bakery_visitors__ = 0

    @classmethod
    async def aopen(cls: Type[T]) -> T:
        """Open bakery."""

        if cls.__bakery_visitors__:
            cls.__bakery_visitors__ += 1
            # no concurrency yet (like aopen/aopen/aopen)
            # anyio lock required (on demand)
            return cls()

        cls.__bakery_visitors__ += 1

        exception: Optional[Union[Exception, BaseException]] = None

        # let's bake all your cakes
        cake: Cakeable[Any]
        for cake in cls.__bakery_items__.values():
            try:
                await bake(cake)
            except (Exception, BaseException) as exc:  # pylint: disable=broad-except
                exception = exc
                logger.error(f'{cake} cannot be baked: {exc}')
                break

        if exception:
            await cls.aclose()
            raise exception

        logger.debug(f"Bakery '{cls}' is opened. Welcome!")
        return cls()

    @classmethod
    async def aclose(
        cls,
        exc_type: Optional[type] = None,
        exc_value: Optional[Exception] = None,
        traceback: Optional[Any] = None,
    ) -> None:
        """Close bakery."""
        cls.__bakery_visitors__ -= 1

        if cls.__bakery_visitors__ > 0:
            logger.debug(f"Bakery '{cls}' is working till the last visitor!")
            return

        exceptions: List[Union[Exception, BaseException]] = []
        item: Cakeable[Any]
        for item in reversed(  # it's important to unbake in reverse order
            # dict views are reversible since 3.8
            # https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects
            list(cls.__bakery_items__.values())
        ):
            try:
                await unbake(item, exc_type, exc_value, traceback)
            except (Exception, BaseException) as exc:  # pylint: disable=broad-except
                exceptions.append(exc)

        logger.debug(f"Bakery '{cls}' is closed.")

        if exceptions:
            # For now raise the first exception occured.
            # Use trio MultiError (inception) on demand
            # https://github.com/python-trio/trio/blob/v0.21.0/trio/_core/_multierror.py#L154
            # or use exception groups (python 3.11)
            # https://peps.python.org/pep-0654/
            raise exceptions[0]



"""Test bakery."""
from inspect import Parameter, Signature
from typing import Any, AsyncIterator, Dict, List, Optional, Type

import pytest
from pytest_mock import MockerFixture

from . import Bakery as _Bakery
from . import (
    Cakeable,
    IngredientsProto,
    bake,
    cake_baking_method,
    cake_ingredients,
    is_cake,
    unbake,
)


Bakery = Type[_Bakery]


def fixture_factory(mocker_name: str) -> Any:
    """Create bakery mock fixtures by mocker name."""

    async def mocker(**kwargs: Any) -> AsyncIterator["BakeryMock"]:
        assert len(kwargs) == 1
        assert list(kwargs.keys())[0].endswith("mocker"), "please, pass mocker fixture."

        mock: BakeryMock = BakeryMock(list(kwargs.values())[0])
        yield mock
        await mock.stopall()

    mocker.__signature__ = Signature(  # type: ignore
        [
            Parameter(
                mocker_name,
                kind=Parameter.KEYWORD_ONLY,
            )
        ]
    )
    return mocker


bakery_mock = pytest.fixture()(fixture_factory("mocker"))
class_bakery_mock = pytest.fixture(scope="class")(fixture_factory("class_mocker"))
module_bakery_mock = pytest.fixture(scope="module")(fixture_factory("module_mocker"))
package_bakery_mock = pytest.fixture(scope="package")(fixture_factory("package_mocker"))
session_bakery_mock = pytest.fixture(scope="session")(fixture_factory("session_mocker"))


class BakeryMock:
    """Toy bakery."""

    def __init__(self, mocker: MockerFixture):
        self._mocker_: MockerFixture = mocker
        self._cake_mocks_: Dict[str, Any] = {}
        self._bakery_: Optional[Bakery] = None
        self._children_: List["BakeryMock"] = []

    def __copy__(self) -> "BakeryMock":
        """Copy bakery mock."""
        mock_copy: BakeryMock = BakeryMock(self._mocker_)
        self._children_.append(mock_copy)
        return mock_copy

    async def _patch_cake(self, *, bakery: Bakery, cake: Cakeable[Any], new_recipe: Any) -> None:
        """Patch cake recipe."""
        assert is_cake(cake)
        assert issubclass(bakery, _Bakery)

        ingredients: IngredientsProto = cake_ingredients(cake)
        if is_cake(new_recipe):
            new_cake: Cakeable[Any] = new_recipe
            self._mocker_.patch.multiple(
                ingredients,
                recipe=cake_ingredients(new_cake).recipe,
                recipe_args=cake_ingredients(new_cake).recipe_args,
                recipe_kwargs=cake_ingredients(new_cake).recipe_kwargs,
                cake_baking_method=cake_baking_method(new_cake),
                result=await bake(new_cake),
            )
        else:
            self._mocker_.patch.multiple(
                ingredients,
                recipe=new_recipe,
                result=new_recipe,
            )

    async def _patch(self, bakery: Bakery) -> None:
        """Start bakery `bakery` patching."""

        if not issubclass(bakery, _Bakery):
            raise ValueError(f"{bakery} is not a Bakery.")

        cake_name: str
        new_recipe: Any

        for cake_name, new_recipe in self._cake_mocks_.items():
            # patch bakery itself
            await self._patch_cake(
                bakery=bakery,
                cake=getattr(bakery, cake_name),
                new_recipe=new_recipe,
            )

        await bakery.aopen()
        self._bakery_ = bakery

    async def patch(self, bakery: Bakery) -> None:
        """Check if bakery set and start patching."""
        if self._bakery_ is not None:
            raise ValueError(f"Old bakery {self._bakery_} is still set. Close it first.")
        return await self._patch(bakery)

    def __call__(self, bakery: Bakery) -> "BakeryMock":
        if self._bakery_ is not None:
            raise ValueError(f"Old bakery {self._bakery_} is still set. Close it first.")
        self._bakery_ = bakery
        return self

    async def __aenter__(self) -> None:
        assert self._bakery_ is not None, "You should set bakery first."
        return await self._patch(self._bakery_)

    async def __aexit__(self, *_args: Any) -> None:
        return await self.reset()

    async def reset(self) -> None:
        """Stop patching."""
        if self._bakery_:
            await self._bakery_.aclose()
            self._bakery_ = None

        self._mocker_.stopall()

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in (
            "_mocker_",
            "_cake_mocks_",
            "_bakery_",
            "_children_",
        ):
            return super().__setattr__(attr, value)

        self._cake_mocks_[attr] = value
        return None

    async def stopall(self) -> None:
        """Reset mocker and clear piece mocks."""
        await self.reset()
        for cake in self._cake_mocks_.values():
            if is_cake(cake):
                await unbake(cake)

        for child in self._children_:
            await child.stopall()

        self._cake_mocks_.clear()
        self._children_.clear()
